#include <tigcclib.h>
#include <estack.h>
#include <args.h>
#include <graph.h>
#include "mlib.h"
#include "anova.h"

void usage() {						// Output usage info
	printf("Usage: canova(Fmat,Rvec,[lbls]) - Perform ANalysis Of VAriance.\n");
	printf("  Fmat     - Matrix of data->factors (normalized)- must have more\n          rows than columns.\n");
	printf("  Rvec     - Column vector of data->results- must have same number\n          of rows as Fmat.\n");
	printf("  lbls     - optional labels for factor columns\n");
	Pause(PROMPT);
	top_estack = (ESI)bottom_estack;
	freeAP();
	exit(1);
}

void InitANVSet(ANVSet **set) {	// Set default values for the elements of the dataset
  *set = (ANVSet *)smalloc(sizeof(ANVSet));

  (*set)->factors = (*set)->results = (*set)->SSf = (*set)->fdof
	= (*set)->Vf = (*set)->Pf = (*set)->lvlsum = (*set)->lvlcnt
	= (*set)->F = (*set)->SSfp = NULL;
  (*set)->T = (*set)->Cf = (*set)->SSt = (*set)->SSe
	= (*set)->ft = (*set)->fe = (*set)->Ve = (*set)->Pe = 0.0;
  (*set)->levels = 0;
  (*set)->clbls = NULL;
  (*set)->poolVars = NULL;
  return;
}

ANVSet *ANVSetup() {				// Do initial program setup
  ANVSet *data;

  InitArgPtr(argptr);	// Set the global stack pointer
  ClrScr();				// Setup the screen
  MoveTo(0,0);
  FontSetSys(USEFONT);
  InitANVSet(&data);	// Setup the data
  printf("%s (%s)\n",COMMENT_PROGRAM_NAME, COMMENT_VERSION_STRING);
  return(data);
}

void GetArgs(ANVSet *data) {		// Get the program arguments from the stack
  register ushort i, j;				// counters
  ushort funcn=5, op=6, farg=7;	// debug arguments
  Matrix *mat;						// return object
  const char *ptr;					// a temporary pointer
  SCR_STATE ss;

  if (GetArgType(argptr) != LIST_TAG) usage();									// Get the factor matrix-- is the opening LIST_TAG missing?
  else {																		//  if not, process the matrix
	GetMatArg(&argptr, &data->factors);
	if (data->factors->n <= 0 || data->factors->m < (data->factors->n+1)) {	// Check factor matrix dimensions
  	  printf("ERROR: Factor matrix wrong size.  Need n%c1 & m>n.\n",0x9c);
  	  printf("       Got a %dx%d matrix.\n",data->factors->m,data->factors->n);
  	  usage();
	}

	data->clbls = (char **)smalloc(sizeof(char *)*(data->factors->m+2));	// Define column label list (room for all factors + ERR + NULL)
	for (i=0; i<data->factors->n; i++) {										// Create default column labels
	  data->clbls[i] = (char *)smalloc(DEFAULT_LBLLEN+1);
	  sprintf(data->clbls[i], "C%c", i+1);
	}
	data->clbls[i++] = ERRSTR;		// Tack on error label
	data->clbls[i] = NULL;			// And terminating NULL

	printf("Got %dx%d factor matrix:\n",data->factors->m,data->factors->n);	// Print factor matrix
	printf("\n");					// Skip the line for the labels for now
	printMatrix(data->factors,3,0,LASTARG);
  }
  if (GetArgType(argptr) != LIST_TAG) usage();									// Get the result vector-- is the opening LIST_TAG missing?
  else {																		//  if not, process the vector
	GetMatArg(&argptr, &data->results);
	if (data->results->n!=1 || data->results->m!=data->factors->m) {			// Check result vector size-- must be (factors->m)-by-1
	  printf("ERROR: Result vector wrong size.  Need %dx1.\n",data->factors->m);
	  printf("       Got a %dx%d matrix.\n",data->results->m,data->results->n);
	  printMatrix(data->results,6,2,LASTARG);
	  usage();
	}
	MoveTo(RVEC_LOC,6);															// Print results vector in the right part of the screen
  	printf("Y =");
  	if (data->clbls) j = 6; else j = 0;
  	prColVecAt(1,RVEC_LOC,12+j,data->results);
  }

  if (GetArgType(argptr) == LIST_TAG) {										// Get factor column labels-- is there a LIST_TAG next?
	argptr--;
	if (GetArgType(argptr) != STR_TAG) {										// Make sure they're strings
	  printf("\nERROR: Column labels must be strings.\n\n");
	  usage();
    }
    for (i=0; i<data->factors->n; i++)
	  if (GetArgType(argptr) == STR_TAG) {										// Make sure the NEXT thing is a string (no numbers, no END_TAG)
		if (strlen(ptr=GetStrnArg(argptr)) > DEFAULT_LBLLEN) {					// If a label is longer than the default limit (from anova.h), reallocate
		  sfree(data->clbls[i]);
		  data->clbls[i] = (char *)smalloc(sizeof(char)*(strlen(ptr)+1));
		}
		strcpy(data->clbls[i],ptr);
	  }
  	if (i!=(data->factors->n)) {													// Did we get enough?
  	  printf("Wrong number of labels (got %d but needed %d).\n", i+1, data->factors->m+1);
  	  usage();
  	}
  }
  SaveScrState(&ss);
  MoveTo(0,0);																	// Now print column labels-- Go to top of screen
  puts("\n\n");																	// Skip the first line.
  for (i=0; i<data->factors->n; i++) 											// Print column labels over the factor matrix
	printf(" %3s ", data->clbls[i]);											// this is very rough-- they won't line up well, but that's OK
  MoveTo(ss.CurX, ss.CurY);
  if (GetArgType(argptr) == POSINT_TAG || GetArgType(argptr) == NEGINT_TAG) {	// DEBUG MODE: Are there debug args?
	funcn = GetIntArg(argptr);		// Specifies matrix function to test
	op = GetIntArg(argptr);			// Specifies op to run within functions that do multiple ops (see mlib.h MatrixOps enum)
	farg = GetIntArg(argptr);		// Specieis scalar argument where appropriate
 	if (funcn > 0) {															// Make sure funcn is positive (must be, since we didn't check the sign!)
	  printf("Debug Mode:\n  funcn=%d, op=%d, farg=%d\n", funcn,op,farg);
	  char ops[5] = "*/+-";
	  switch (funcn) {
	  	case 1:										// New matrix (op=0)
	  	mat = newMat(3,3,farg);
	  	printf("newMat(3,3,%d):\n", farg);
	  	printMatrix(mat,LASTARG);
	  	break;

	  	case 2:										// elem-wise op
	  	mat = eWiseOp(data->factors,op,data->results);
	  	printf("eWiseOp([],%c,[]):\n",ops[op]);
	  	if (mat) printMatrix(mat,LASTARG);
	  	else printf("Returned NULL.\n");
	  	break;

	  	case 3:										// scalar op
	  	mat = sOpMat(farg,op,data->factors);
	  	printf("sOpMat(%d,%c,[]):\n",farg,ops[op]);
	  	if (mat) printMatrix(mat,LASTARG);
	  	else printf("Returned NULL.\n");
	  	break;

	  	case 4:										// column sum
	  	mat = colSum(data->factors);
	  	printf("colSum([]):\n");
	  	if (mat) printMatrix(mat,LASTARG);
	  	else printf("Returned NULL.\n");
	  	break;
	  }
	  printf("\nDone.\n");
	  Pause(PROMPT);
	  top_estack = (ESI)bottom_estack;
	  freeAP();
	  return;
	}
  }
  top_estack = argptr;															// Done getting args-- this should be redundant, but just to be safe...
}

void showInputs(ANVSet *data) {		// Display the input data
  register ushort i;				// counter

  ClrScr();
  MoveTo(0,0);																	// Now print column labels-- Go to top of screen
  printf("Got %dx%d factor matrix:\n",data->factors->m,data->factors->n);		// Print factor matrix
  for (i=0; i<data->factors->n; i++) 											// Print column labels over the factor matrix
	printf(" %3s ", data->clbls[i]);											// this is very rough-- they won't line up well, but that's OK
  puts("\n");
  printMatrix(data->factors,3,0,LASTARG);

  MoveTo(RVEC_LOC,6);															// Print results vector in the right part of the screen
  printf("Y =");
  prColVecAt(1,RVEC_LOC,2*FONTHEIGHT,data->results);
  Pause(PROMPT);
}

void calcANOVA(ANVSet *data) {		// Perform ANOVA
  register ushort i,j,k;			// counters
  Matrix *mat, *tmat;				// temp matrices to use during calculations

  // Step 1. Total of data->results						(data->T)
  for (i=0, data->T=0.0; i < data->results->m; data->T += data->results->dat[i++]);

  // Step 2. Correction factor 							(data->Cf)
  data->Cf = data->T*data->T/data->factors->m;		// Cf = T^2/n = T*T/factors->m

  // Step 3. Total sum of squares 						(data->SSt)
  mat = eWiseOp(data->results,MOP_MULT,data->results);
  for (i=0, data->SSt=0.0; i < mat->m; data->SSt += mat->dat[i++]);
  data->SSt -= data->Cf;
  sfree(mat);
  printf("Results: T = %f     Cf = %f     SSt = %f\n", data->T, data->Cf, data->SSt);

  // Step 4. Factor sums of squares						(data->SSf)
  mat = colMax(data->factors);
  mat = trn(mat);
  tmat = colMax(mat);
  data->levels = (ushort)(tmat->dat[0]);
  sfree(tmat); sfree(mat);
  data->lvlsum = newMat(data->levels,data->factors->n,0.0);
  data->lvlcnt = newMat(data->levels,data->factors->n,0.0);
  for (j=0; j < data->factors->n; j++)
    for (k=0; k < data->levels; k++)
	  for (i=0; i < data->factors->m; i++)
		if (data->factors->dat[i*data->factors->n+j]==(k+1)) {
		  data->lvlsum->dat[k*data->factors->n+j] += data->results->dat[i];
		  data->lvlcnt->dat[k*data->factors->n+j]++;
		}
  mat = eWiseOp(data->lvlsum,MOP_MULT,data->lvlsum);	// mat = data->lvlsum.^2		A1^2
  tmat = eWiseOp(mat,MOP_DIV,data->lvlcnt);			// tmat = mat./data->lvlcnt   A1^2/NA1
  sfree(mat);
  mat = colSum(tmat);									// Si = Sum(Aj^2/NAj)
  sfree(tmat);
  data->SSf = sOpMat((-1.0*data->Cf),MOP_ADD,mat);	// Si = Sum(Aj^2/NAj) - data->Cf
  sfree(mat);
  printf("SSf = ");
  printMatrix(data->SSf,7,4,LASTARG);
  for (i=0, data->SSe=0.0; i < data->SSf->n; data->SSe += data->SSf->dat[i++]);
  data->SSe = data->SSt - data->SSe;
  data->SSf->dat[data->factors->n] = data->SSe;		// HACK

  // Step 5. Factor degrees of freedom					(data->fdof)
  data->ft = data->factors->m - 1;
  data->fdof = newMat(1,data->factors->n,data->levels-1);
  data->fe = data->ft - (data->levels-1)*data->factors->n;		//mat[0];  This assumes all factors have the same number of levels
  data->fdof->dat[data->factors->n] = data->fe;  		// HACK
  printf("fe = %f     f = ", data->fe);
  printMatrix(data->fdof,3,1,LASTARG);

  // Step 6. Variance									(data->Vf)
  data->Vf = eWiseOp(data->SSf,MOP_DIV,data->fdof);
  printf("Vf = ");
  printMatrix(data->Vf,7,4,LASTARG);
  data->Vf->specials = (char **)smalloc(sizeof(char *)*(data->factors->n+1));
  for (i=0; i <= data->factors->n; data->Vf->specials[i++] = NULL);
  if (abs(data->fe)<epsilon) {
  	data->Ve = -10.0;	// <0 indicates undefined for variance
  	printf("Ve indeterminate.\n");
	data->Vf->specials[data->factors->n] = NASTR;
  }	else {
    data->Ve = data->SSe/data->fe;
	data->Vf->dat[data->factors->n] = data->Ve;  // HACK
    printf("Ve = %f\n", data->Ve);
  }

  // Step 7. Relative contributions						(data->Pf)
  data->Pf = sOpMat(100.0/data->SSt,MOP_MULT,data->SSf);
  data->Pe = 0.0;
  data->Pf->dat[data->factors->n] = data->Pe;			// HACK
  printf("Pf = ");
  printMatrix(data->Pf,7,4,LASTARG);
  data->SSf->m=data->SSf->n;	// Make these into column vectors instead of the row vectors they are now.
  data->SSf->n=1;
  data->fdof->m=data->fdof->n;
  data->fdof->n=1;
  data->Vf->m=data->Vf->n;
  data->Vf->n=1;
  data->Pf->m=data->Pf->n;
  data->Pf->n=1;

  // Setup the empty vars
  data->F = newMat(data->factors->n,1,0.0);										// Make empty matrices whose specials arrays
  data->SSfp = newMat(data->factors->n,1,0.0);									//  hold the dash marker.
  data->F->specials = (char **)smalloc(sizeof(char *)*(data->factors->n+1));
  data->SSfp->specials = (char **)smalloc(sizeof(char *)*(data->factors->n+1));
  for (i=0; i <= data->factors->n; i++)
    data->SSfp->specials[i] = data->F->specials[i] = &DASH[0];
  Pause(PROMPT);
}

void printANOVA(ANVSet *data) {		// Print the ANOVA results table
  register ushort i,j;				// counters
  static char *buffer = NULL;		//

  if (!buffer) buffer = (char *)smalloc(sizeof(char)*128);
  // Use global vars shared between printANOVA() and fmtANOVA()
  mlist = (Matrix **)smalloc(sizeof(Matrix *)*7);
  footer = (float **)smalloc(sizeof(float *)*7);
  // Clumsily initialize with {NULL, data->fdof, data->SSf, data->Vf, data->F, data->SSfp, data->Pf}
  mlist[0]=NULL; mlist[1]=data->fdof; mlist[2]=data->SSf; mlist[3]=data->Vf; mlist[4]=data->F; mlist[5]=data->SSfp; mlist[6]=data->Pf;
  // Clumsily initialize with {NULL, &(data->ft), &(data->SSt), NULL, NULL, NULL, NULL};
  footer[0]=NULL; footer[1]=&(data->ft); footer[2]=&(data->SSt); footer[3]=footer[4]=footer[5]=footer[6]=NULL;

  ClrScr();
  j=pos[0];
  MoveTo(pos[0],0);
  puts("Fac");
  MoveTo(pos[0],FONTHEIGHT);
  puts("-----");
  for (i=0; i <= data->factors->n; i++) {		// Print factor label column by hand
	if (i < data->factors->n) {				// all the regular entries
	  sprintf(buffer, "%d:%s", i+1, data->clbls[i]?data->clbls[i]:"");
	  prAlnStr(pos[0], (i+2)*FONTHEIGHT, widths[0]*FONTWIDTH, AL_LEFT, buffer, 0);
	}
  }
  if (data->poolVars && data->poolVars[0])	// If pooled variables exist, output the error line label
	prAlnStr(pos[0], (i+1)*FONTHEIGHT, widths[0]*FONTWIDTH, AL_LEFT, data->clbls[data->factors->n], 0);
  puts("\n-----\nTOTAL:");

  for (i = 1; i < 7; i++) {					// Print columns 2-7 (as specified in anova.h)
	if (data->poolVars && data->poolVars[0]) (mlist[i]->m)++;
	prFmtVecAt(pos[i], 0, mlist[i], &lbls[i][0], widths[i], footer[i]);
	if (data->poolVars && data->poolVars[0]) (mlist[i]->m)--;
  }
  // Other data
  printf("St = %f     Se = %f     ft = %f     fe = %f     Ve = ", data->SSt, data->SSe, data->ft, data->fe);
  if (data->Ve < 0.0) printf("undef\n");
	else printf("%f\n",data->Ve);
  Pause(PROMPT);
}

int askOptions(char **rSpec, ANVSet *data) {	// Ask the user what to do with the data
  int i;							// counter
  int choice;						// menu choice
  ushort ans;						// character entered
  static char *buffer = NULL;		// Input buffer
  char *ptra, *ptrb;				// pointers for string manipulation
  SCR_STATE ss;

  // Initialize the buffers
  if (!buffer) {buffer = (char *)smalloc(16); buffer[0]=0;}
  if (!(*rSpec)) *rSpec = (char *)smalloc(64);
  (*rSpec)[0]=0;
  printf("Option Menu:\n");													  // Display the menu
  printf(" 1) Display input data\n");
  printf(" 2) Display current ANOVA table\n");
  printf(" 3) Save current ANOVA results to data file\n");
  printf(" 4) Exit now and return these results as answer\n");
  printf(" 5) Select factors for pooling and reanalyze current inputs\n");
  printf(" 6) Quit (discard result data)");
  printf("\nSelect (1-6): ");
  SaveScrState (&ss);
  putchar('_');
  for (choice = 0; ((choice = ngetchx()) < '1') || (choice > '6'); );	// Loop until a key between 1 and 6 is pressed
  MoveTo (ss.CurX, ss.CurY);
  printf("%c\n", choice);
  switch(choice) {															// Process the keypress
	case '3':																// 3. Save to file
	do {																	//    Loop until we get either a usable filename or an empty string
	  if (strlen(*rSpec)) puts("Filename invalid.  Try again.\n");
	  puts("Enter filename: ");
	  InputStr(*rSpec, 18);
	} while (strlen(*rSpec) && TokenizeSymName(*rSpec, TSF_FULLY_QUALIFIED) == NULL_INDEX);
	GetSymstrArg(top_estack); 												//    Pop the symname off the stack (we checked them with a function that pushes the result onto the stack if successful
	if (!strlen(*rSpec)) return(0);										//    If we got here because of an empty string, just reprint the menu
	break;

	case '5':
	if (!data->poolVars) {													//    If pooling list has not been initialized, do so now
	  data->poolVars = (ushort *)smalloc(sizeof(ushort)*(data->factors->n+1));
	  data->poolVars[0] = 0;
	}
	if (data->poolVars[0]) {												//    If the pooling list is non-empty, build a string of the factors already pooled
	  sprintf(*rSpec, "%u", data->poolVars[0]);
	  for (i=1; data->poolVars[i]; i++) {
		sprintf(buffer, ",%u", data->poolVars[i]);
		strcat(*rSpec, buffer);
	  }
	} else (*rSpec)[0] = 0;
	printf("List factors to be pooled (e.g. 1,4,5) from 1-%u: ", data->factors->n);
	InputStr(*rSpec, 60);
	if (!strlen(*rSpec)) {													//    If answer is empty, do plain ANOVA
	  data->poolVars[0] = 0;	// Set poolVar list to empty
	  return(-1);				// Return indicator
	} else {
	  ptra = strcpy(buffer, *rSpec);
	  printf("Pooling factors: ");
	  for (i=0; (ptrb = strtok(ptra,",")); i++) {							//    Parse the reply string (of factors to pool) and store the results
		ans = (ushort)atoi(ptrb);
		ptra = NULL;
		if (ans < 1 || ans > data->factors->n) { 							//    Is a given factor valid?
		  printf("\n  ERROR: Entry %d (%d) is invalid.  Must be from 1 to %d.",
			i+1, data->poolVars[i], data->factors->n);
		  Pause(PROMPT);
		  return(0);
		}
		data->poolVars[i] = ans;
		printf("%d:%s ",ans, data->clbls[ans-1]);
	  }
	  data->poolVars[i] = 0;	// Add the terminator to the list
	  puts("\n");
	  for (i=0; data->poolVars[i]; i++)									//    Check the complete list of pooled factors
		if (data->Pf->dat[data->poolVars[i]-1] > 5.0 ) {					//    If any have more than 5% Pf, issue a warning.
		  printf(" *** WARNING *** Contribution of factor %d (%s) is %f%%\n",
			data->poolVars[i], data->clbls[data->poolVars[i]-1],
			data->Pf->dat[data->poolVars[i]-1]);
		  printf("        This is more than 5%% %c Factor may be significant!\n", 0x16);
		}
	  Pause(PROMPT);
	}
	break;

	case '6':																// 6. Quit
	top_estack = (ESI)bottom_estack;										//    Mark the stack as empty
	freeAP();																//    free smalloc()'d memory
	exit(0);
  }
  return(choice-48);														// return the keypress minus the ASCII value of '0'
}

void fmtANOVA(ANVSet *data) {		// Format the output data and push it onto the stack
  register ushort i,j;				// counters
  char *cmdstr, *curln;				// buffer space for strings

  // These gotta be this big to be safe
  cmdstr = (char *)smalloc(4096);
  curln = (char *)smalloc(1024);

  printf("Formatting result data... ");
  push_END_TAG();
  for (j=6; j; j--) {							// Output last 6 columns (to push onto stack in reverse order)
	sprintf(cmdstr,"{\"%s\"", lbls[j]);		// Column label as text
	for (i=0; i < data->factors->n; i++) {	// For all factors, output either the numerical or the specials entry
	  if (mlist[j]->specials					// If the specials array exists,
			&& mlist[j]->specials[i]			//   AND the ith entry is non-NULL,
			&& mlist[j]->specials[i][0]) {	//   AND the ith entry is non-empty
	    sprintf(curln, ",\"%s\"", mlist[j]->specials[i]+1);	// Use that entry, bypassing the alignment token (which is the first character)
	  } else {
		sprintf(curln, ",%f", mlist[j]->dat[i]);				// Otherwise, use the ith numerical value
	  }
	  strcat(cmdstr,curln);
	}
	strcat(cmdstr,"}");
	push_parse_text(cmdstr);					// Push onto the stack
  }
  sprintf(cmdstr,"{\"Fct\"");					// Add the factor label column
  for (i=0; i<data->factors->n; i++) {		//   For each factor, add the right label
    if (data->clbls && data->clbls[i]) {		//   Check to make sure the labels list is setup (it should ALWAYS be, by now)
	  sprintf(curln,",\"%s\"",data->clbls[i]);	//   If so, use the label
	} else {
	  sprintf(curln,",%d",i+1);				//   Otherwise, just put the factor number
	}
    strcat(cmdstr,curln);
  }
  strcat(cmdstr, "}");
  push_parse_text(cmdstr);
  sprintf(cmdstr, "{\"Col\"");					// Add the factor number column.  If there were no labels AND the default labels
  for (i=0; i<data->factors->n; i++) {		//    somehow didn't get setup right, the factors column will be identical to this one.
    sprintf(curln,",%d",i+1);
    strcat(cmdstr,curln);
  }
  strcat(cmdstr, "}");
  push_parse_text(cmdstr);
  printf("Done.\n");
}

void outputANOVA(char *outfn) {		// Output data on stack to var or as program return value
  char *ptr;		// pointer for string manipulation

  if (strlen(outfn?outfn:"")) {					// Save to data variable and exit
	TokenizeSymName(outfn, TSF_FULLY_QUALIFIED);	// Convert to fully-qualified
	for (ptr = (char *)(GetSymstrArg(top_estack))-1; ptr[0]; ptr--);// printf("T:%s\n", ptr);
	TokenizeSymName(outfn, TSF_FULLY_QUALIFIED);	// Convert to fully-qualified
	cmd_newdata(top_estack);
	printf("Data saved to %s\n",ptr+1);
	top_estack = (ESI)bottom_estack;
	Pause(PROMPT);
	return;
  }
  push_LIST_TAG();									// Return data to stack
  push_quantum(TRANSPOSE_TAG);						// transpose it, though
  NG_approxESI(top_estack);							// process it
  freeAP();
  exit(0);
}

void calcAPool(ANVSet *data) {		// Recalculate with pooled factors
  register ushort i,j;				// Counters
  ushort facPooled;					// Is the current factor pooled?
  static char *buffer=NULL;		// string buffer

  ClrScr();
  MoveTo(0,0);
  printf("%s (%s) - Pooling Analysis\n",COMMENT_PROGRAM_NAME, COMMENT_VERSION_STRING);

  // Setup buffer  and specials for pooled variables IFF not already done
  if (!buffer) buffer = (char *)smalloc(sizeof(char)*64);
  if (!data->fdof->specials) data->fdof->specials = (char **)smalloc(sizeof(char *)*(data->factors->n+1));
  if (!data->SSf->specials) data->SSf->specials = (char **)smalloc(sizeof(char *)*(data->factors->n+1));
  if (!data->Pf->specials) data->Pf->specials = (char **)smalloc(sizeof(char *)*(data->factors->n+1));
  for (i=0; i < data->factors->n; i++) data->Vf->specials[i] = data->fdof->specials[i]
  	= data->SSf->specials[i] = data->Pf->specials[i] = NULL;

  // Perform ANOVA calculations
  data->SSe = 0;																// Step 1a
  data->fe = 0;																	// Step 2a
  printf("\nPooled Factors: ");
  for (i=0; data->poolVars[i]; i++) {											// For each pooled variable
	j = data->poolVars[i]-1;													// j = the ith pooled variable
    printf("%d:%s ", j+1, data->clbls[j]);
	data->SSe += data->SSf->dat[j];											// Step 1b  (SSe += SSfj)
	data->fe += data->fdof->dat[j];											// Step 2b  (Fe += fj)

	// And Vf="POOLED", and ()'s for f and S for pooled factors
	data->Vf->specials[j] = POOLSTR;
	sprintf(buffer, "R(%f)", data->fdof->dat[j]);
	data->fdof->specials[j] = (char *)smalloc(sizeof(char)*(strlen(buffer)+1));
	strcpy(data->fdof->specials[j], buffer);
	sprintf(buffer, "R(%f)", data->SSf->dat[j]);
	data->SSf->specials[j] = (char *)smalloc(sizeof(char)*(strlen(buffer)+1));
	strcpy(data->SSf->specials[j], buffer);
	data->Pf->specials[j] = data->F->specials[j] = data->SSfp->specials[j] = &DASH[0];
  }
  data->SSf->dat[data->factors->n] = data->SSe;
  data->SSf->specials[data->factors->n] = NULL;
  data->fdof->dat[data->factors->n] = data->fe;
  data->fdof->specials[data->factors->n] = NULL;
  if (abs(data->fe) > epsilon) {												// Step 3 - Is fe ~= 0.0?
	data->Ve = data->SSe/data->fe;												// No => calculate Ve
	data->Vf->dat[data->factors->n] = data->Ve;								// Store in the error slot at the bottom of the Vf column
	data->Vf->specials[data->factors->n] = NULL;								// Unset the corresponding Vf specials entry (causing the number to print)
  } else {
	data->Ve = -10;															// Yes => Set Ve to a sentinel value
	data->Vf->specials[data->factors->n] = NASTR;								// Set the specials entry at the bottom of the Vf column to 'n/a'
  }
  data->Pe = 100.0;
  printf("\n\nSignificant Factors: ");
  for (i=0; i < data->factors->n; i++) {										// For each variable
	for (j=facPooled=0; !facPooled && data->poolVars[j]; j++)
	  if (i == data->poolVars[j]-1) facPooled = 1;							// Is the var pooled?
	if (!facPooled) {															// NO => For significant factors only...
	  printf("%d:%s ", i+1, data->clbls[i+1]);
	  data->F->dat[i] = data->Vf->dat[i]/data->Ve;							// Step 4
	  data->F->specials[i] = NULL;											//   unset specials entry
	  data->SSfp->dat[i] = data->SSf->dat[i] - data->Ve*data->fdof->dat[i];	// Step 5
	  data->SSfp->specials[i] = NULL;											//   unset specials entry
	  data->Pf->dat[i] = 100.0*data->SSfp->dat[i]/data->SSt;					// Step 6
	  data->Pf->specials[i] = NULL;											//   unset specials entry
	  data->Pe -= data->Pf->dat[i];											// Pe -= Pfi
	}
  }
  data->Pf->dat[data->factors->n] = data->Pe;									// Pe is what's leftover.  Store it in the Pf error slot
  data->Pf->specials[data->factors->n] = NULL;								// unset the specials entry for the error slot
  puts("\n\n");
  Pause(PROMPT);
}


