#include "mlib.h"
#include "anova.h"

void GetMatArg(ESI *aptr, Matrix **mat) {		// Get a matrix argument from the stack
  register ushort i,j;
  ushort m=0, n=0;
  int sign=1,tint,num,denom;
  ESI tmpptr = *aptr;

  *mat = NULL;	// just in case
  if (GetArgType(tmpptr) != LIST_TAG) return;
  tmpptr--;
  for (i=1; (GetArgType(tmpptr)!=END_TAG) && i<10; i++) {// Count rows.
	tmpptr--;	// Advance past row list tag
	for (j=1; GetArgType(tmpptr)!=END_TAG && j<10; j++) {
	  SkipArg(tmpptr);
	  if (j > n) n = j;
	}
	tmpptr--;	// Skip row end tag
	if (i > m) m = i;
  }
  *mat = (Matrix *)smalloc(sizeof(Matrix));
  (*mat)->dat = (float *)smalloc(sizeof(float) * m * n);
  (*mat)->m = m;
  (*mat)->n = n;

  tmpptr = *aptr - 1;
  for (i=0; i<m; i++) {
	tmpptr--;		// Advance past row list tag
	for(j=0; j<n; j++) {
	  switch (GetArgType(tmpptr)) {
		case NEGINT_TAG:
		sign=-1;
		case POSINT_TAG:
		tint = GetIntArg(tmpptr);
		(*mat)->dat[i*n+j] = (float)(sign*tint);
		sign=1;
		break;

		case FLOAT_TAG:
		(*mat)->dat[i*n+j] = GetFloatArg(tmpptr);
		break;

		case NEGFRAC_TAG:
		sign=-1;
		case POSFRAC_TAG:
		num = sign*GetIntArg(tmpptr);
		sign=1;
		tmpptr++;
		denom = GetIntArg(tmpptr);
		(*mat)->dat[i*n+j] = ((float)num)/((float)denom);
		break;

		default:
		(*mat)->dat[i*n+j] = 0.0;
		SkipArg(tmpptr);
	  }
	}
	tmpptr--;	// row end tag
  }
  *aptr=tmpptr-1;
  return;
}

void printMatrix(Matrix *mat, ... ) { 			
  // int fw=5, int prec=2, int zpad=0
  register ushort i,j;
  int argval=0, fw=5, prec=2, zpad=0;
  char fmt[11];
  va_list vargs;

  va_start(vargs, mat);
  for (i=0; i < 3; i++) {
	argval = va_arg(vargs, int);
	if (argval == LASTARG) break;
	switch(i) {
	  case 0:
	  fw = argval;
	  break;

	  case 1:
	  prec = argval;
	  break;

	  case 2:
	  zpad = argval;
	}
  }
  va_end(vargs);
  sprintf(&fmt[0], " %%%s%d.%df ",zpad?"0":"",fw,prec);
	for (i=0; i < mat->m; i++) {
	  putchar((mat->m==1)?'[':'|');
	  for (j=0; j<mat->n; j++) printf(fmt,mat->dat[i*mat->n+j]);
	  printf("%c\n", (mat->m==1)?']':'|');
	}
  return;
}

void prColVecAt(ushort delim, ushort x, ushort y, Matrix *mat, ...) {
  register ushort i;
  ushort col = 1;
  va_list vl;

  if (!mat) return;
  if (mat->n > 1) {
	va_start(vl, mat);
	col = va_arg(vl,ushort);
	va_end(vl);
  }
  col--;
  if (x > 236-(delim?4:0)) x = 236-(delim?4:0);
  for (i=0; i<mat->m; i++) {
	if (y > 124) y = 124;
	MoveTo(x,y);
	if (delim) printf("|%f|",mat->dat[i*mat->n + col]);
	else printf("%f",mat->dat[i*mat->n + col]);
	y+=6;
  }
  MoveTo(0,y);	// Effective CR
}

Matrix *newMat(ushort m, ushort n, float defaultVal) {
  register ushort i,j;
  Matrix *mat;

  if (m==0 || n==0) return(NULL);
  mat = (Matrix *)smalloc(sizeof(Matrix));
  mat->dat = (float *)smalloc(sizeof(float) * (m * n + 2));	// Extra for storing error
  mat->dat[m*n] = defaultVal;// HACK
  mat->m = m;
  mat->n = n;
  mat->specials = NULL;

  for (i=0; i<m; i++)
	for (j=0; j<n; j++)
	  mat->dat[i*n+j] = defaultVal;
  return(mat);
}

Matrix *eWiseOp(Matrix *mat1, ushort op, Matrix *mat2) {	// Elementwise ops of matrices
  register ushort i,j;
  ushort m,n;
  Matrix *rmat;

  if (!mat1 || !mat2) return(NULL);
  if (mat1->m!=mat2->m || mat1->n!=mat2->n || (op>MOP_SUBTR)) return(NULL);	// Must have same dimensions
  m = mat1->m;
  n = mat1->n;
  rmat = newMat(m, n, 0.0);
	for (i=0; i<m; i++)
	  for (j=0; j<n; j++)
		switch (op) {
		  case MOP_MULT:
		  rmat->dat[i*n+j] = mat1->dat[i*n+j] * mat2->dat[i*n+j];
		  break;

		  case MOP_DIV:
		  rmat->dat[i*n+j] = mat1->dat[i*n+j] / mat2->dat[i*n+j];
		  break;

		  case MOP_ADD:
		  rmat->dat[i*n+j] = mat1->dat[i*n+j] + mat2->dat[i*n+j];
		  break;

		  case MOP_SUBTR:
		  rmat->dat[i*n+j] = mat1->dat[i*n+j] - mat2->dat[i*n+j];
		  break;
		}
  return(rmat);
}

Matrix *sOpMat(float x, ushort op, Matrix *mat) {		// Scalar ops on matrix
  register ushort i,j;
  ushort m,n;
  Matrix *rmat;

  if (!mat || op>MOP_SUBTR) return(NULL);
  m = mat->m;
  n = mat->n;
  rmat = newMat(m,n,0.0);
  for (i=0; i<m; i++)
	for (j=0; j<n; j++)
	  switch (op) {
		case MOP_MULT:
		rmat->dat[i*n+j] = x * mat->dat[i*n+j];
		break;

		case MOP_DIV:
		rmat->dat[i*n+j] = x / mat->dat[i*n+j];
		break;

		case MOP_ADD:
		rmat->dat[i*n+j] = x + mat->dat[i*n+j];
		break;

		case MOP_SUBTR:
		rmat->dat[i*n+j] = x - mat->dat[i*n+j];
		break;
	  }
  return(rmat);
}

Matrix *colSum(Matrix *mat) {	// Column sums of matrix
  register ushort i,j;
  Matrix *rmat;

  if (!mat) return(NULL);
  rmat = newMat(1, mat->n, 0.0);
  for (j=0; j<mat->n; j++) {		// For each column
	rmat->dat[j] = 0;
	for (i=0; i<mat->m; i++)		// Go down it and add
	  rmat->dat[j] += mat->dat[i*mat->n+j];
  }
  return(rmat);
}

Matrix *colMax(Matrix *mat) {	// Columnwise max of matrix
  register ushort i,j;
  Matrix *rmat;

  if (!mat) return(NULL);
  rmat = newMat(1, mat->n, 0.0);
  for (j=0; j<mat->n; j++) {		// For each column
	rmat->dat[j] = 0;
	for (i=0; i<mat->m; i++)		// Go down it and add
	  if (rmat->dat[j] < mat->dat[i*mat->n+j])
		rmat->dat[j] = mat->dat[i*mat->n+j];
  }
  return(rmat);
}

Matrix *trn(Matrix *mat){
  register ushort i,j;
  Matrix *rmat;

  if (!mat) return(NULL);
  rmat = newMat(mat->n, mat->m,0.0);
  for (i=0; i < mat->m; i++) {
	for (j=0; j < mat->n; j++) rmat->dat[j*rmat->n + i] = mat->dat[i*mat->n + j];
  }
  return(rmat);
}

void InputStr(char *buffer, ushort maxlen) {  // Taken verbatim from the TIGCC help (http://tigcc.ticacl.org/)
  short key;
  ushort i = strlen(buffer);
  SCR_STATE ss;
  SaveScrState (&ss);
  do {
    MoveTo (ss.CurX, ss.CurY);
    printf ("%s_  ", buffer);
    // Note that two spaces are required only if the F_4x6 font is used
    key = ngetchx ();
    if (key >= ' ' && key <= '~' && i < maxlen)
	  buffer[i++] = key;
    else if (key == KEY_BACKSPACE && i)
	  i--;
    buffer[i] = 0;
  } while (key != KEY_ENTER);
  MoveTo (ss.CurX, ss.CurY);
  puts(buffer);
  puts("\n");
}

void *smalloc(unsigned long len) {
  static ushort blocks = 0;										// Causes smalloc on 1st call
  ushort i;
  void *ptr = malloc(len);

  if (!allocptrs) blocks = 0;			// In case sfreeAP() was called in the middle of a prog
  for (i = 0; blocks && allocptrs[i]; i++);								// Find the end marker
  if ((!blocks) || !((i+1)%16))	{										// Expand every 16th time
  	allocptrs = (void **)realloc(allocptrs, sizeof(void *)*(i+16+3));	// Really only need +1, but for added safety...
    blocks++;															// Increment the block counter
  }
  allocptrs[i] = ptr;													// Save the pointer
  allocptrs[i+1] = NULL;												// tack on a new end marker
  if (!ptr) {
	printf("\n\nFAILED malloc(%ld)%c%d!!!\n", len, 0x16, i+1);
	ngetchx();
  }
  return(ptr);
}

void sfree(void *ptr) {
  register ushort i;

  for (i = 0; allocptrs[i] && allocptrs[i]!=ptr; i++);	// Find the given pointer.  NULL marks the end.
  if (!allocptrs[i]) return;								// If we get NULL, we were asked to free a ptr not in the list.
  free(allocptrs[i]);										// free it
  for ( ; (allocptrs[i] = allocptrs[i+1]); i++);			// Slide the rest down
  return;
}

void freeAP(void) {
  register ushort i;

  if (!allocptrs) return;								// Allocated memory list does not exist yet.
  for (i = 0; allocptrs[i]; free(allocptrs[i++]));	// free() everything in the list
  free(allocptrs);										// free() the list itself
  allocptrs = NULL;									// Mark it empty
}

void prAlnColVecAt(ushort delim, ushort colPW, ushort align, ushort x, ushort y, Matrix *mat, ...) {
  register ushort i,j;
  static char buffer[80];

  ushort col=1;														// Column to use out of a rectangular matrix
  va_list vl;

  if (!mat) return;
  
  if (mat->n > 1) {												// If the matrix is rectangular, which column do we use?
	va_start(vl, mat);
	col = va_arg(vl,ushort);
	va_end(vl);
  }
  col--;
  if (x > 236-(delim?4:0)) x = 236-(delim?4:0);	// sanity check
  for (i=0; i < mat->m; i++) {									// Loop through all the rows of the vector/matrix
	if (y > 124) y = 124;		// sanity check
	if ((mat->specials) && (mat->specials[i*mat->n + col])) {	// Let an entry in the specials array override a numerical value
	  if (strlen(mat->specials[i*mat->n + col])>1) {
	    j = 1;
	    switch (mat->specials[i*mat->n + col][0]) {
	      case 'C': align=AL_CENTER; break;
	      case 'L': align=AL_LEFT; break;
	      case 'R': align=AL_RIGHT; break;
	      case 'D': align=AL_DECIMAL; }
	  } else j=0;
	  strcpy(buffer, (mat->specials[i*mat->n + col])+j);
	} else {
	  if (delim) sprintf(buffer, "|%f|", mat->dat[i*mat->n + col]);
	  else sprintf(buffer, "%f", mat->dat[i*mat->n + col]);
	}
	prAlnStr(x,y,colPW,align,buffer,'#');
	y += FONTHEIGHT;												// Move output point down one line
  }
  MoveTo(0,y);	// Effective CR when done
}

void prFmtVecAt(ushort x, ushort y, Matrix *mat, char *label, ushort colWidth, float *footer) {
  static char *fstr = NULL, line[] = "--------------------------";
  ushort colPW = colWidth*FONTWIDTH;						// Line pixel height and column pixel width

  if (!fstr) fstr = (char *)smalloc(sizeof(char)*64);	// Allocate space for fstr (only the first time around)
  prAlnStr(x,y,colPW,AL_CENTER,label,(char)0);				// Output column label
  strcpy(fstr, line);
  prAlnStr(x,y+FONTHEIGHT,colPW,AL_CENTER,fstr,(char)0);	// Output top-of-column rule
  if (!mat) return;										// If matrix is empty, then column header is all we need
  prAlnColVecAt(0,colPW,1,x,y+2*FONTHEIGHT,mat);			// Output right-aligned vector
  strcpy(fstr, line);
  prAlnStr(x,y+(mat->m+2)*FONTHEIGHT,colPW,AL_CENTER,fstr,(char)0);	// Output bottom-of-column rule
  if (footer) {						// Output footer data (e.g. total of column, average, whatever)
	sprintf(fstr,"%f", *footer);
  	prAlnStr(x,y+(mat->m+3)*FONTHEIGHT,colPW,AL_RIGHT,fstr,'#');	// Output right-aligned footer
  }
  MoveTo(0,y+(mat->m+4)*FONTHEIGHT);
}

void prAlnStr(ushort x, ushort y, ushort cellPW, ushort align, char *str, char tChar) {
  static ushort strPW = 0, slen;	// String pixelwidth and length

  if ((x+cellPW) >= 240) {			// Shrink the effective cellwidth to the right edge of the screen
    if (x >= 236) return;
    cellPW = 239-x;
  }
  while ((strPW=DrawStrWidth(str, USEFONT)) > cellPW) {	// truncate string, if necessary
	slen = strlen(str);
	str[--slen] = 0;
	if (tChar) str[--slen] = tChar;
  }
  switch (align) {					// Move to coords determined by alignment
    default:
	case AL_LEFT:					// Left align
	MoveTo(x,y);
	break;

	case AL_DECIMAL:				// Align on decimal point (not yet implemented)
	case AL_RIGHT:					// Right align
	MoveTo(x+cellPW-strPW,y);
	break;

	case AL_CENTER:
	MoveTo(x+(cellPW-strPW)/2,y);	// Centered
	break;
  }
  puts(str);						// Output the string
}











